@@ -2,22 +2,23 @@ mod debug;
 mod primitives;
 mod raycast;
 
-pub use crate::debug::*;
-pub use crate::primitives::*;
+pub use crate::{debug::*, primitives::*};
 
 use crate::raycast::*;
-use bevy::math::Vec3A;
-use bevy::render::render_resource::PrimitiveTopology;
 use bevy::{
+    core::FloatOrd,
     ecs::schedule::ShouldRun,
+    math::Vec3A,
     prelude::*,
     render::{
         camera::Camera,
         mesh::{Indices, Mesh, VertexAttributeValues},
+        render_resource::PrimitiveTopology,
     },
-    tasks::prelude::*,
+    tasks::ComputeTaskPool,
 };
 use std::{
+    collections::BTreeMap,
     marker::PhantomData,
     sync::{Arc, Mutex},
 };
@@ -332,8 +333,8 @@ pub fn build_rays<T: 'static + Send + Sync>(
 #[allow(clippy::type_complexity)]
 pub fn update_raycast<T: 'static + Send + Sync>(
     // Resources
-    pool: Res<ComputeTaskPool>,
     meshes: Res<Assets<Mesh>>,
+    task_pool: Res<ComputeTaskPool>,
     // Queries
     mut pick_source_query: Query<&mut RayCastSource<T>>,
     culling_query: Query<
@@ -363,9 +364,9 @@ pub fn update_raycast<T: 'static + Send + Sync>(
             let ray_cull = info_span!("ray culling");
             let raycast = info_span!("raycast");
 
+            let ray_cull_guard = ray_cull.enter();
             // Check all entities to see if the source ray intersects the AABB, use this
             // to build a short list of entities that are in the path of the ray.
-            let ray_cull_guard = ray_cull.enter();
             let culled_list: Vec<Entity> = culling_query
                 .iter()
                 .filter_map(
@@ -376,31 +377,33 @@ pub fn update_raycast<T: 'static + Send + Sync>(
                             } else {
                                 visibility.is_visible
                             };
-                        if !should_raycast {
-                            None // Exit early for entities that we can skip
-                        } else if let Some(aabb) = bound_vol {
-                            if let Some([_, far]) =
-                                ray.intersects_aabb(aabb, &transform.compute_matrix())
-                            {
-                                if far < 0.0 {
-                                    None // AABB is behind the ray
+                        if should_raycast {
+                            if let Some(aabb) = bound_vol {
+                                if let Some([_, far]) =
+                                    ray.intersects_aabb(aabb, &transform.compute_matrix())
+                                {
+                                    if far >= 0.0 {
+                                        Some(entity)
+                                    } else {
+                                        None
+                                    }
                                 } else {
-                                    Some(entity)
+                                    None
                                 }
                             } else {
-                                None // Ray does not intersect the AABB
+                                Some(entity)
                             }
                         } else {
-                            Some(entity) // Has no AABB
+                            None
                         }
                     },
                 )
                 .collect();
             drop(ray_cull_guard);
-            let picks = Arc::new(Mutex::new(Vec::new()));
 
+            let picks = Arc::new(Mutex::new(BTreeMap::new()));
             mesh_query.par_for_each(
-                &pool,
+                &task_pool,
                 32,
                 |(mesh_handle, simplified_mesh, transform, entity)| {
                     if culled_list.contains(&entity) {
@@ -412,20 +415,15 @@ pub fn update_raycast<T: 'static + Send + Sync>(
                             if let Some(intersection) =
                                 ray_intersection_over_mesh(mesh, &transform.compute_matrix(), &ray)
                             {
-                                picks.clone().lock().unwrap().push((entity, intersection));
+                                picks.lock().unwrap().insert(
+                                    FloatOrd(intersection.distance()),
+                                    (entity, intersection),
+                                );
                             }
                         }
                     }
                 },
             );
-            let mut picks = Arc::try_unwrap(picks).unwrap().into_inner().unwrap();
-            picks.sort_by(|a, b| {
-                a.1.distance()
-                    .partial_cmp(&b.1.distance())
-                    .unwrap_or(std::cmp::Ordering::Equal)
-            });
-
-            pick_source.intersections = picks;
         }
     }
 }
@@ -466,7 +464,7 @@ pub fn ray_intersection_over_mesh(
                 vertex_positions,
                 vertex_normals,
                 ray,
-                Some(&vertex_indices.iter().map(|x| *x as u32).collect()),
+                Some(vertex_indices),
             ),
             Indices::U32(vertex_indices) => ray_mesh_intersection(
                 mesh_to_world,
@@ -477,7 +475,27 @@ pub fn ray_intersection_over_mesh(
             ),
         }
     } else {
-        ray_mesh_intersection(mesh_to_world, vertex_positions, vertex_normals, ray, None)
+        ray_mesh_intersection(
+            mesh_to_world,
+            vertex_positions,
+            vertex_normals,
+            ray,
+            None::<&Vec<u32>>,
+        )
+    }
+}
+
+pub trait IntoUsize: Copy {
+    fn into_usize(self) -> usize;
+}
+impl IntoUsize for u16 {
+    fn into_usize(self) -> usize {
+        self as usize
+    }
+}
+impl IntoUsize for u32 {
+    fn into_usize(self) -> usize {
+        self as usize
     }
 }
 
@@ -487,7 +505,7 @@ pub fn ray_mesh_intersection(
     vertex_positions: &[[f32; 3]],
     vertex_normals: Option<&[[f32; 3]]>,
     pick_ray: &Ray3d,
-    indices: Option<&Vec<u32>>,
+    indices: Option<&Vec<impl IntoUsize>>,
 ) -> Option<Intersection> {
     // The ray cast can hit the same mesh many times, so we need to track which hit is
     // closest to the camera, and record that.
@@ -512,15 +530,15 @@ pub fn ray_mesh_intersection(
         // chunk of three indices are references to the three vertices of a triangle.
         for index in indices.chunks(3) {
             let tri_vertex_positions = [
-                Vec3A::from(vertex_positions[index[0] as usize]),
-                Vec3A::from(vertex_positions[index[1] as usize]),
-                Vec3A::from(vertex_positions[index[2] as usize]),
+                Vec3A::from(vertex_positions[index[0].into_usize()]),
+                Vec3A::from(vertex_positions[index[1].into_usize()]),
+                Vec3A::from(vertex_positions[index[2].into_usize()]),
             ];
             let tri_normals = vertex_normals.map(|normals| {
                 [
-                    Vec3A::from(normals[index[0] as usize]),
-                    Vec3A::from(normals[index[1] as usize]),
-                    Vec3A::from(normals[index[2] as usize]),
+                    Vec3A::from(normals[index[0].into_usize()]),
+                    Vec3A::from(normals[index[1].into_usize()]),
+                    Vec3A::from(normals[index[2].into_usize()]),
                 ]
             });
             let intersection = triangle_intersection(
